 /*
  * tcpsyn.h
  *
  * go forth and syn no more.
  */

/*
 * notes on this data structure
 *	fins: a large amount of syns and a small amount of fins is a NEGATIVE indicator
 *		we either have an attacker or an ineffective protocol.
 *		equal amounts of syns/fins may be regarded as a good sign.
 *		NEGATIVE if small, and/or syns significantly greater.
 *
 *	rst: non-trivial rsts, are of course a bad sign. 
 *		NEGATIVE.
 *
 *	tcpfrom/tcpto are a measure of work.  small tcpto is a bad sign.
 *		work. NEGATIVE if from greater than to.
 *
 *	icmpcount: icmp errors that we see in the real world.  also a negative indicator.
 *		NEGATIVE.
 */
#define WEIGHTCOUNTER	float

/* info flag bits for this ip src address */
#define P2PBITS		0x2ff		/* do not include port flags */

#define IRCJOINS	0x1		/* irc message type */
#define IRCPINGS	0x2		/* irc message type */
#define IRCPONGS	0x4		/* irc message type */
#define IRCPRIVMSG	0x8		/* irc message type */
#define BITTORRENT	0x10		/* peer connection initiation */
#define GNUTELLA	0x20		/* peer connection initiation */
#define KAZAA		0x40		/* kazaa connect initiation */
#define EDONKEY 	0x80 		/* more of a guess */
#define MORPHEUS	0x200		/* morpheus/limewire */
#define WEBSERVER	0x400		/* src port is 80 or 443 */
#define EMAIL		0x800		/* any dst port is 25 */

#define MAXPORTS			10
struct synnode {
	unsigned long ipsrc; 		/* index */
	unsigned long syncount;		/* syns sent by ip src */
	unsigned long fincount;		/* fins returned to ip src */
	unsigned long finssent;		/* fins sent by ip src */
	unsigned long rstcount;		/* resets returned to ip src */
	unsigned long sacount;		/* number of syn+ack pairs sent by ip src */
	unsigned long tcpfrom;		/* total tcp packets sent from ipsrc */
	unsigned long tcpto;		/* total tcp pkts sent to ip src */
	unsigned long icmpcount;	/* total icmp errors sent back to the ip src */ 
	unsigned long flags;		/* flags related to ip src */
	unsigned long emulehits;	/* count of emule guesses */
	WEIGHTCOUNTER work;		/* work weight + a little fudge*/
	WEIGHTCOUNTER ww;		/* work weight with no fudge at all */
	int pcount;			/* number of ports counted */
	int dports[MAXPORTS];   	/* max dst ports */
	unsigned long dpcounts[MAXPORTS];   /* associated dst port counters */
	struct synnode *next;
};

/*
#define HSIZE 1031 
*/
#define THSIZE 16403
struct synlist{
	unsigned long entries;
	struct synnode *last;
	struct synnode *hashlist[THSIZE];
};

static unsigned long e_hash(unsigned char *, int size);

/**************************************************/
/* function prototypes */

void init_synlist(struct synlist *l);

void search_srcsyn(struct t_entry *te, struct synlist *l, int saflag, int ircresult, int icmpflag);

void destroy_synlist(struct synlist *l);
void print_synlist(FILE *fd, struct synlist *l, int count, int wflag, 
	char *wormfile, int ipaflag, unsigned long usnet, 
	unsigned long usmask, int scanflag, struct slist *isl, struct slist *tsl);
