#! /usr/bin/perl

#
# daily.pl
#
# syntax:
#	# daily.pl base_log_dir
# e.g.,
#	# daily.pl /home/mrourmon/logs
#
# run at around midnight to start log directory over for next day.
# run from crontab.
# 0	0	*	*	*	root	/home/mrourmon/bin/daily.pl /home/mrourmon/logs
#
# All log dirs are found under a base log directory with the following dir scheme:
#
# /home/mrourmon/logs
#	mon.lite
#	tworm
#	topn
#	portreport  - backed backend worm reports in portsig format
#	p2preport  - backed backend p2p reports in portsig format
#	rawirc     - raw frontend irc reports 
#	irc        - baked backend irc reports 
#	udpreport  - baked backend udp error reports
#	rawp2p     - raw frontend p2p logs
#
#	each directory above has Sun Mon Tue ... Sat daily directories
#
# This perl script is called once at midnight to prepare for the next
# day's logging.
#
# it's job is to zero out the daily (Sun Mon Tue Wed Thu Fri Sat)
# directory in the base logging directory for all 3 of the above.
#
# for the topn logging directory it also creates a symlink
# to a cur for the current day.
#
# Assumption: output_dir has subdirs for days of week.  Mon, Tue, etc.
# Thus output format is:  log/Mon/date.mon.lite
# form is thus:
# Sun Mon Tue Wed Thu Fri Sat 
#
# Note daily sym links are NOT used by any software.  Just for
# humans.

$debug = 0;
$argc=$#ARGV+1;

if ($argc != 1) {
	print "syntax: # daily.pl output_dir\n";
	print "example: # daily.pl /home/mrourmon/logs\n";
	exit(1);
}
use File::Basename();

use Time::Local;
# comment in for ctime, note that it overrides localtime
use Time::localtime;

my $outputdir = $ARGV[0];

################################################### 
$start_time=`date`;
$start_time=~ s/\n//g;		# zap the newline at the end of the date
@time = split(/ /, $start_time);
$dow = $time[0];

chdir($outputdir);

# mon.lite daily prep
system("rm -fr mon.lite/$dow"); 
system("rm -fr mon.lite/mon_today"); 
system("mkdir mon.lite/$dow"); 
system("ln -s $dow mon.lite/mon_today");

# tworm daily prep
system("rm -fr tworm/$dow"); 
system("rm -fr tworm/tworm_today"); 
system("mkdir tworm/$dow"); 
system("ln -s $dow tworm/tworm_today");

# topn daily prep
system("rm -fr $dow"); 
system("rm -f topn_today"); 
system("mkdir $dow"); 
system("ln -s $dow topn_today");

# portreport daily prep
system("rm -fr portreport/$dow"); 
system("rm -f portreport/portreport_today"); 
system("mkdir portreport/$dow"); 
system("ln -s $dow portreport/portreport_today");

# p2preport daily prep
system("rm -fr p2preport/$dow"); 
system("rm -f p2preport/p2preport_today"); 
system("mkdir p2preport/$dow"); 
system("ln -s $dow p2preport/p2preport_today");

# rawirc daily prep
system("rm -fr rawirc/$dow"); 
system("rm -f rawirc/rawirc_today"); 
system("mkdir rawirc/$dow"); 
system("ln -s $dow rawirc/rawirc_today");

# irc daily prep
system("rm -fr irc/$dow"); 
system("rm -f irc/irc_today"); 
system("mkdir irc/$dow"); 
system("ln -s $dow irc/irc_today");

# udpreport daily prep
system("rm -fr udpreport/$dow"); 
system("rm -f udpreport/udpreport_today"); 
system("mkdir udpreport/$dow"); 
system("ln -s $dow udpreport/udpreport_today");

# rawp2p daily prep
system("rm -fr rawp2p/$dow"); 
system("rm -f rawp2p/rawp2p_today"); 
system("mkdir rawp2p/$dow"); 
system("ln -s $dow rawp2p/rawp2p_today");

exit(0);

# end of main


