#! /usr/bin/perl

#
# topipa.pl - weights IP address network activity
#
# For syntax:
#	topipa.pl [ -h ]  [ -n num ]  [ -s "A,B,C,D" ] 
#	    [ -d <directory> ] [ -f "mon.lite_file0 ... monlite_fileN" ]
#	    [ -dw <directory> ] [ -fw "tcpworm_file0 ... tcpworm_fileN" ]
#	   [ -od <directory> ] [ -html ] [ -debug ] [ -o <outfile> ] 
#	   [ -report ]
#
# input file(s):
#	mon.lite  - file which is output of ourmon front-end binary
#	   	note: mon.lite specifies individual filters by name
#
#	tcpworm.txt
#
# outputs:	   
#	topipa.html (default)
#	host html files (optional)
#
# Modified from omupdate.pl
#
# tuples used:  (all of them)
#       SYNTUPLE
#       UDPTUPLE
#       SCANTUPLE
#       FLOWTUPLE
#
# derek@cs.pdx.edu 03/13/2005
#

########################################################

#
# Each element consists of a counter incremented when the key ip 
# address is observed in the corresponding field of a mon.lite
# file.
#
# %hosts hash of hashes
#
# %hosts = (
#	 ipaddr0 => {
#		 count => 0,
#		 topn_tcp => 0,
#		 topn_udp => 0,
#		 topn_icmp => 0,
#		 syn_list => 0,
#		 tcpworm => 0,
#		 ip_scan => 0,
#		 tcp_scan => 0,
#		 udp_scan => 0,
#		 icmp_error => 0,
# 		 udp_error => 0,
#		 flows => 0,
#		 metrics => 0,
#		 scans => 0,
#		 score => 0,
#		 data => {
#			topn_tcp =>
#				flowid => {
#					start_time,
#					stop_time,
#					tuple =	{ concatonated tuples array },
#					{ topn_tcp fields array },
#				}
#				...
#			topn_udp =>
#				flowid => {
#					start_time,
#					stop_time,
#					tuple =	{ concatonated tuples array },
#					{ topn_udp fields array },
#				}
#				...
#			topn_icmp.
#				flowid => {
#					start_time,
#					stop_time,
#					tuple =	{ concatonated tuples array },
#					{ topn_icmp fields array },
#				}
#				...
#			}
#			...
#		}
#	 },
#	 ipaddr1 => { 
#		 ...
#	 {
#
#
# It didn't start this way and I never imagined it would grow like 
# this. At first all it did was count. The assosiative array expaneded 
# as processing was added. Seems to work anyway.
#

use vars qw/ %opt /;
use Getopt::Long;

my %hosts = ();
my %perIPFlags = ();
my %perIPPorts = ();
my @fields = qw/topn_tcp topn_udp topn_icmp syn_list udp_error 
	icmp_error ip_scan tcp_scan udp_scan tcpworm/ ;
my @catagories = qw/flows metrics scans worm/;
my @flows = qw/topn_tcp topn_udp topn_icmp/ ;
my @metrics = qw/syn_list icmp_error udp_error/ ;
my @scans = qw/ip_scan tcp_scan udp_scan/ ;
my @worm = qw/tcpworm/;
my @ports = qw/6881 27000/;

my $hostwebdir = "../web.pages/hosts";
my $directory="";
my $scoreFlows = 1;
my $scoreMetrics = 1;
my $scoreScans = 1;
my $scoreWorms = 1;
my $numlines = 100;
my @FILES = "";
my @WORMFILES = "";
my $TRUE = 1;
my $FALSE = 0;
my $DEBUG = $FALSE;
my $REPORT = $FALSE;
my $TOP = $FALSE;
my $delimiter = "\n";
my $hostdirectory = $hostwebdir;;
my $tablelines = 5;
my $htmlfile = "../web.pages/topipa.html";
my $tophtml = "../web.pages/top.html";
my $start_time;
my $stop_time;
my $timestamp;
my $date;
my $color1 = #e5e5e5;
my $color2 = #c0c0c0;
my $colorDebug = #ff0000;


#
# catagory index
#
my $flows_index = 3;
my $metrics_index = 3;
my $scans_index = 3;



#
# tuple step values
#
my $flows_step = 2;
my $step_icmp_error = 8;
my $step_udp_error = 10;
my $step_ip_scan = 2;
my $step_tcp_scan = 2;
my $step_udp_scan = 2;
my $step_syn_list = 15;
my $step_worm = 6;



#
# a key to mon.lite entries. subject to change according to jim's whimsy
# headers sans source ip
#
my %fieldheaders = (
	topn_tcp => [ qw/source_port destination_ip.port byte_count/ ],
	topn_udp => [ qw/source_port destination_ip.port byte_count/ ],
	topn_icmp => [ qw/source_port destination_ip.port byte_count/ ],
	syn_list => [ qw/syns synack fin_count fin_sent rst tcpfrom
		tcpto icmp_errors flags irc emulehits ipds tcpds ports/ ], 
	icmp_error => [ qw/total tcp udp ping icmp_unreachable 
		icmp_redirect icmp_ttl_exceeded/ ],
	udp_error => [ qw/weight sent returned icmp_unreachable
		ping ipds tcpds port_count ports/ ],
	ip_scan => [ qw/count/ ],
	tcp_scan => [ qw/count/ ],
	udp_scan => [ qw/count/ ],
	tcpworm => [ qw/syncount sacount fincount finssent rstcount 
		tcpfrom tcpto icmpcount flags emulehits ipds tcpds 
		pcount ports/ ],
);

my %portFlags = (
	6881 => "P?",
	27000 => "G?",
);

my %months = (
	Jan => 1, Feb => 2, Mar => 3, Apr => 4, May => 5, Jun => 6,
	Jul => 7, Aug => 8, Sep => 9, Oct => 10, Nov => 11, Dec => 12,
);



#
# subroutine list
#
sub print_report;
sub print_html_report;
sub print_tcpworm_summary;
sub print_scan_summary;
sub print_metrics_summary; 
sub print_flows_table;
sub print_top;
sub print_html; 
sub printUsageAndExit;
sub flows_update;
sub scans_update;
sub metrics_update;
sub timestamp_update; 
sub parseAppFlags;



#
# BEGIN
#


if ( (scalar @ARGV) < 1 ) { printUsageAndExit(); }

#
# process comand line options
#
GetOptions(
	"h" => \$opt{help},
	"help" => \$opt{help},
	"debug" => \$opt{debug},
	"report" => \$opt{report},
	"top" => \$opt{top},
	"n:i" => \$opt{n},
	"s:s" => \$opt{s},
	"d:s" => \$opt{d},
	"f:s" => \$opt{f},
	"fw:s" => \$opt{ftcpworm},
	"dw:s" => \$opt{dtcpworm},
	"do:s" => \$opt{directory},
	"o:s" => \$opt{htmlfile},
	"wp:s" => \$opt{hostwebdir},
);

if ( $opt{'help'} ) {
	printUsageAndExit();
}
if ( $opt{'htmlfile'}) {
	$htmlfile = $opt{'htmlfile'};
}
if ( $opt{'n'} ) {
	$numlines = $opt{'n'};
}
if ( $opt{'d'} ) {

	my $size;

	$directory = $opt{'d'};

	#
	# read all files in a directory into the array and
	# prepend directory path to filename
	#
	opendir (DIR, $directory) or 
	    die "can't open $directory directory: $!\n";
	@FILES = readdir DIR;
	closedir DIR;
	$size = scalar @FILES;
	while ($size > -1) {
		# prepend directory path to files
		if ( -f "$directory/$FILES[$size]" ) {
			$FILES[$size] = $directory.'/'.$FILES[$size];
		}
		$size--;
	}
} 
if ( $opt{'dtcpworm'} ) {

	my $size;

	$directory = $opt{'dtcpworm'};

	#
	# read all files in a directory into the array and
	# prepend directory path to filename
	#
	opendir (DIR, $directory) or 
	    die "can't open $directory directory: $!\n";
	@WORMFILES = readdir DIR;
	closedir DIR;
	$size = scalar @WORMFILES;
	while ($size > -1) {
		# prepend directory path to files
		if ( -f "$directory/$WORMFILES[$size]" ) {
			$WORMFILES[$size] = $directory.'/'.$WORMFILES[$size];
		}
		$size--;
	}
}
if ( $opt{'f'} ) {
	@FILES = split / /, $opt{'f'};
}
if ( $opt{'ftcpworm'} ) {
	@WORMFILES = split / /, $opt{'ftcpworm'};
}
if ( $opt{'s'} ) {
	($scoreFlows, $scoreMetrics, $scoreScans, $scoreWorms) = split /,/, $opt{'s'};
}
if ( $opt{'html'} ) { $HTML = $TRUE; }
if ( $opt{'debug'} ) { $DEBUG = $TRUE; }
if ( $opt{'report'} ) { $REPORT = $TRUE; }
if ( $opt{'top'} ) { $TOP = $TRUE; }
if ( $opt{'directory'} ) { $hostdirectory = $opt{'directory'}; }
if ( $opt{'hostwebdir'} ) { $hostwebdir = $opt{'hostwebdir'}; }


#
# process mon.lite files
#
while ( $_ = shift @FILES ) {

	my $weekday;
	my $month;
	my $day;
	my $time;
	my $timezone;
	my $year;
	my $fname;
	my @FILE;

	next if ( $_ eq "\." || $_ eq "\.\." );

	open(FILE, $_) || die "can't open $_ file: $!\n";
	@FILE = <FILE>;
	close(FILE);

	$fname = `basename $_`;
	chomp $fname;

	($weekday, $month, $day, $time, $timezone, $year) =
		split /_/, $fname;

	if ( defined $month ) {
		$year =~ s/\.\w+\.\w+\.\w+$//;
		$timestamp = $year . "-" . $months{$month} . "-" . 
			$day . "-" . $time;
	} else {
		$timestamp = "0000-00-00-00:00:00";
	}

	if ($start_time gt $timestamp || ! defined $start_time ){
		    $start_time = $timestamp;
	}

	if ($stop_time lt $timestamp || ! defined $stop_time ){
		    $stop_time = $timestamp;
	}


	$phrases = join("<br>",@FILE);

	#  fetch each line from datafile, line is separated by "\n"

	@phrases = split(/$delimiter/,$phrases);

	# walk phrases and process on line by line basis
	# one subroutine per filter type

	foreach my $elem (@phrases) {

		# get rid of blanks, return, line breaks

		$elem =~ s/\n//g;
		$elem =~ s/<br>//g;
		$elem =~ s/ //g;	

		# save each item to array @items_in_eachline, from each line,

		@items_in_eachline = split(/:/,$elem);

		if ($items_in_eachline[0] eq "topn_tcp") {
			flows_update("topn_tcp", @items_in_eachline);
		}
		elsif ($items_in_eachline[0] eq "topn_udp") {
			flows_update("topn_udp", @items_in_eachline);
		}
		elsif ($items_in_eachline[0] eq "topn_icmp") {			
			flows_update("topn_icmp", @items_in_eachline);
		}
		elsif ($items_in_eachline[0] eq "syn_list") {
			metrics_update("syn_list", @items_in_eachline);
		}
		elsif ($items_in_eachline[3] eq "worm_log") {
#			metrics_update("worm", @items_in_eachline);
		}
		elsif ($items_in_eachline[0] eq "ip_scan") {
			scans_update("ip_scan", @items_in_eachline);
		}
		elsif ($items_in_eachline[0] eq "ip_portscan") {
#			scans_update("L4_scan", @items_in_eachline);
		}
		elsif ($items_in_eachline[0] eq "tcp_portscan") {
			scans_update("tcp_scan", @items_in_eachline);
		}
		elsif ($items_in_eachline[0] eq "udp_portscan") {
			scans_update("udp_scan", @items_in_eachline);
		}
		elsif ($items_in_eachline[0] eq "icmperror_list") {
			metrics_update("icmp_error", @items_in_eachline);
		}
		elsif ($items_in_eachline[0] eq "udperror_list") {
			metrics_update("udp_error", @items_in_eachline);
		}
	}  # end loop
} # end file loop



#
# process tcpworm.txt files
#
while ( $_ = shift @WORMFILES ) {

	my $weekday;
	my $month;
	my $day;
	my $time;
	my $timezone;
	my $year;
	my $host;
	my $fname;
	my $port;

	next if ( $_ eq "\." || $_ eq "\.\." );

	open FILE, $_ or die "can't open $_ file: $!\n";
	@WORM=<FILE>;
	close FILE;

	$fname = `basename $_`;
	chomp $fname;

	($weekday, $month, $day, $time, $timezone, $year) =
		split /_/, $fname;

	if ( defined $month ) {
		$year =~ s/\.\w+\.\w+$//;
		$timestamp = $year . "-" . $months{$month} . "-" . 
			$day . "-" . $time;
	} else {
		$timestamp = "0000-00-00-00:00:00";
	}

	if ($start_time gt $timestamp || ! defined $start_time ){
		    $start_time = $timestamp;
	}

	if ($stop_time lt $timestamp || ! defined $stop_time ){
		    $stop_time = $timestamp;
	}

	foreach my $elem (@WORM) {
		@items_in_eachline = split(/\:/,$elem);
		$host = $items_in_eachline[0];
		$hosts{$host}{'tcpworm'}++;
		$hosts{$host}{'count'}++;
		shift @items_in_eachline;

		foreach my $item ( @{ $fieldheaders{'tcpworm'} } ) {
			my $i = shift @items_in_eachline;
			push @{ $hosts{$host}{'data'}{'tcpworm'}{$item} },
				$i;
		}

		# calculate work from first array element
		$work = int ( ( 
			$hosts{$host}{'data'}{'tcpworm'}{'syncount'}[0] +
			$hosts{$host}{'data'}{'tcpworm'}{'finssent'}[0] +
			$hosts{$host}{'data'}{'tcpworm'}{'rstcount'}[0] ) /
			(
			$hosts{$host}{'data'}{'tcpworm'}{'tcpfrom'}[0] +
			$hosts{$host}{'data'}{'tcpworm'}{'tcpto'}[0] 
			) * 100 );

		push @{ $hosts{$host}{'data'}{'tcpworm'}{'work'} }, $work;
		timestamp_update($host,"tcpworm");

		# initialize flags if necessary
		if ( ! defined $perIPFlags{$host} ) {
			$perIPFLags{$host} = 0;
		}
	
		$perIPFlags{$host} = $perIPFlags{$hosts} | 
			$hosts{$host}{'data'}{'tcpworm'}{'flags'}[0];

		foreach $elem ( @ports ) {
			foreach $port ( split /,/, @{ $hosts{$host}{'data'}{'tcpworm'}{'ports'}[0] } ) {
				if ( $elem == $port ) {
					$perIPPorts{$host}{$elem}++;
				}
				$perIPPorts{$host}{$elem}++;
			}
		}

	}

}




#
# calculate sum of all fields for validity
#
if ( $DEBUG ) {
	foreach $i ( keys %hosts ) {
		foreach $j ( @fields ) {
			$hosts{$i}{'sum'} += $hosts{$i}{$j};
		}
	}
}


# 
# calculate score:
#
#	 A * flows + B * metrics + C * scans
#
foreach $i ( keys %hosts ) {

	foreach $j ( qw/flows metrics scans/ ) {
		$hosts{$i}{$j} = 0;
	}
	foreach $j ( @flows ) {
		$hosts{$i}{'flows'} += $hosts{$i}{$j};
	}
	foreach $j ( @metrics ) {
		$hosts{$i}{'metrics'} += $hosts{$i}{$j};
	}
	foreach $j ( @scans ) {
		$hosts{$i}{'scans'} += $hosts{$i}{$j};
	}

	$hosts{$i}{'score'} = ($scoreFlows * $hosts{$i}{'flows'}) + 
		($scoreMetrics * $hosts{$i}{'metrics'}) + 
		($scoreScans * $hosts{$i}{'scans'}) +
		($scoreWorms * $hosts{$i}{'tcpworm'}) ;
}

#
# show results
#
if ( $REPORT ) {
	print_report()
}
if ( $TOP ) {
	print_top();
}

print_html();
exit 0;

#
# END
#


##
#
# subroutines
#
##


#
# html host report output
#
sub print_report() {

	my $n = $numlines;
	my $date = localtime time;

	# sort in decreasing score
	foreach $i (sort { $hosts{$b}{'score'} <=> $hosts{$a}{'score'} } 
	  keys %hosts) {
		if ( $n < 0 ) { last; }
		print_html_report($i);
		$n--;
	}

} # end sub report_html


#
#
#
sub print_html_report() {
	my $host = shift;
	my $field;

	my $fname = $hostdirectory . "/" . $host . "\.html";
	open (HFILE, "> $fname" ) or die "can't open $host: $!\n";


print HFILE <<FIN
<html>
<style type=\"text/css\">
body, table, th, td {
    font-family: Arial, Helvetica, sans-serif;
}
th { text-align: center }
td { text-align: right; width:\"10%\" }
</style>
<body bgcolor=\"#ffffff\">
<head>
<title> $host </title>
</head>
<h3 align=\"center\">$host</h3>
FIN
;

	foreach $field ( @fields ) {
		my $line = 0;

		# skip empty fields
		next if ( $hosts{$host}{$field} < 1 );
		print HFILE "<h3 align=\"left\"> $field </h3>\n\n";

		#
		# print tables for flows only.
		#
		if ( grep { /$field/m } @flows ) {
			print_flows_table($host,$field);
		}

		#
		# print summary for scans
		#
		elsif ( grep { /$field/m } @scans ) {
			print_scan_summary($host,$field);	
		}

		#
		# print summary for metrics
		#
		elsif ( grep { /$field/m } @metrics ) {
			print_metrics_summary($host,$field);
		}

		#
		# print summary for tcpworm
		#
		elsif ( $field eq "tcpworm" ) {
			print_tcpworm_summary($host);
		}
	}

	print HFILE "<p align=\"left\">$date</p>\n";
	print HFILE "</body>\n";
	print HFILE "</html>\n";
	close HFILE;

} # end sub print_html_report


sub print_tcpworm_summary() {
	my $host = shift;
	my $mean = 0;
	my $median = 0;
	my $variance = 0;
	my @temp;
	my $elem;
	my $i;
	my $appFlags;
	my %data;
	my %cols;
	my %ports;
	my $elem;
	my $lines=0;

	foreach $elem ( @{ $hosts{$host}{'data'}{'tcpworm'}{'work'} } ) {
		$mean += $elem;
	}	
	if ( $mean > 0 ){
		$mean = int ( $mean / scalar @{ $hosts{$host}{'data'}{'tcpworm'}{'work'} });
	}

	foreach $elem ( @{ $hosts{$host}{'data'}{'tcpworm'}{'work'} } ) {
		$i = $elem - $mean;
		$variance += $i * $i;
	}
	
	if ( $variance > 0 ) {
		$variance = int sqrt ( $variance/ $hosts{$host}{'tcpworm'} );
	}
	@temp = sort { $a <=> $b } @{ $hosts{$host}{'data'}{'tcpworm'}{'work'} };
	$median = $temp[ int ( (scalar @temp + 1) /2 ) ];

	$appFlags = parseAppFlags( $perIPFlags{$host} );

	foreach $elem ( @ports ) {
		if ( $perIPPorts{$host}{$elem} > 0 ) {
			$appFlags .= $portFlags{$elem};
		}
	}

	if ( $appFlags eq "" ) {
		$appFlags = "None";
	}

print HFILE <<FIN
<p> 
Count: $hosts{$host}{'tcpworm'}, 
Avg. Work: $mean, 
Max. Work: $temp[$#temp],
App Flags: $appFlags
</p>
<table border="0" align="center">
<thead>
<tr align="center" bgcolor="#c0c0c0">
<th> mean </th>
<th> median </th>
<th> variance </th>
<th> max </th>
<th> min </th>
<th> start </th>
<th> stop </th>
</thead>

<tr bgcolor="#e5e5e5">
<td> $mean </td>
<td> $median </td>
<td> $variance </td>
<td> $temp[$#temp] </td>
<td> $temp[0] </td>
<td> $hosts{$host}{'data'}{'tcpworm'}{'start_time'} </td>
<td> $hosts{$host}{'data'}{'tcpworm'}{'stop_time'} </td>
</tr>
</table>

FIN
;

	# print html table header
	print HFILE "<p></p>\n";
	print HFILE "<table border=\"0\" align=\"center\">\n";
	print HFILE "<thead>\n";
	print HFILE "<tr align=\"center\" bgcolor=\"#c0c0c0\">\n";
	print HFILE "<th> work </th>\n";

	foreach $elem ( @{ $fieldheaders{'tcpworm'} } ) {
       		print HFILE "<th> $elem </th>\n";
		if ( $elem eq "flags" ) {
			print HFILE "<th> appFlags </th>\n";
		}
       	}
	# add start/stop time to header
	print HFILE "</tr>\n";
	print HFILE "</thead>\n";
	print HFILE "<tbody>\n";

	# copy arrays
	foreach $elem ( @{ $fieldheaders{'tcpworm'} } ) {
		@{ $data{$elem} } = 
			 @{ $hosts{$host}{'data'}{'tcpworm'}{$elem} };
	}

# DEBUG
#	foreach $elem ( @{ $fieldheaders{'tcpworm'} } ) {
#		print "$elem		@{ $data{$elem} }\n";
#	}

	$lines = 0;
	while( $lines < $tablelines ) {

		if ( $lines++ % 2 == 0 ) {
	       		print HFILE "<tr bgcolor=\"#e5e5e5\">\n"
		} else {
       			print HFILE "<tr bgcolor=\"#c0c0f0\">\n"
		}

		$elem = shift @{ $hosts{$host}{'data'}{'tcpworm'}{'work'} };

		if ( defined $elem ) {
			print HFILE "<td> $elem </td>\n";
		}

		foreach  $elem ( @{ $fieldheaders{'tcpworm'} } ) {
			$i = shift @{ $data{$elem} };
			next if ( ! defined $i );
			print HFILE "<td> $i </td>\n";
			if ( $elem eq "flags" ) {
				$appFlags = parseAppFlags($i);
				print HFILE "<td> $appFlags </td>\n";
			}
		}
		print HFILE "</tr>\n";
	}
}




#
# print scan summary
#
sub print_scan_summary() {
	my $host = shift;
	my $scan = shift;
	my $mean = 0;
	my $median = 0;
	my $variance = 0;
	my @temp;
	my $elem;
	my $i;

	foreach $elem ( @{ $hosts{$host}{'data'}{$scan}{'count'} } ) {
		$mean += $elem;
	}	
	if ( $mean > 0 ){
		$mean = int $mean / scalar @{ $hosts{$host}{'data'}{$scan}{'count'} };
	}

	foreach $elem ( @{ $hosts{$host}{'data'}{$scan}{'count'} } ) {
		$i = $elem - $mean;
		$variance += $i * $i;
	}
	
	if ( $variance > 0 ) {
		$variance = int sqrt ( $variance/ $hosts{$host}{$scan} );
	}
	@temp = sort { $a <=> $b } @{ $hosts{$host}{'data'}{$scan}{'count'} };
	$median = $temp[ int ( (scalar @temp + 1) /2 ) ];

print HFILE <<FIN
<h3 align="left"> $field </h3>
<p> Count $hosts{$host}{$scan} </p>
<table border="0" align="center">
<thead>
<tr align="center" bgcolor="#c0c0c0">
<th> mean </th>
<th> median </th>
<th> variance </th>
<th> max </th>
<th> min </th>
<th> start </th>
<th> stop </th>
</thead>

<tr bgcolor="#e5e5e5">
<td> $mean </td>
<td> $median </td>
<td> $variance </td>
<td> $temp[$#temp] </td>
<td> $temp[0] </td>
<td> $hosts{$host}{'data'}{$scan}{'start_time'} </td>
<td> $hosts{$host}{'data'}{$scan}{'stop_time'} </td>
</tr>
</table>

FIN
;

}


#
# print metrics summary
#
sub print_metrics_summary() {
	my $host = shift;
	my $metric = shift;
	my $field;
	my @temp;
	my $mean = 0;
	my $median = 0;
	my $variance = 0;
	my $elem;
	my $i = 0;;

	if ($metric eq "syn_list") {
		@temp = @{ $hosts{$host}{'data'}{$metric}{'syns'} };
	} elsif ($metric eq "icmp_error") {
		@temp = @{ $hosts{$host}{'data'}{$metric}{'total'} };
	} elsif ($metric eq "udp_error") {
		@temp = @{ $hosts{$host}{'data'}{$metric}{'weight'} };
	}

	foreach $elem (@temp) {
		$mean += $elem
	}
	if ( $mean > 0 ) {
		$mean = int $mean / scalar @temp;
	}

	foreach $elem (@temp) {
		$i = $elem - $mean;
		$variance += $i * $i;
	}

	if ( $variance > 0 ) {
		$variance = int sqrt ( $variance/ $hosts{$host}{$metric} );
	}
	@temp = sort { $a <=> $b } @temp;

	$median = $temp[ int ( (scalar @temp + 1) /2 ) ];

print HFILE <<FIN
<h3 align="left"> $field </h3>
<p> Count $hosts{$host}{$metric} </p>
<table border="0" align="center">
<thead>
<tr align="center" bgcolor="#c0c0c0">
<th> mean </th>
<th> median </th>
<th> variance </th>
<th> max </th>
<th> min </th>
<th> start </th>
<th> stop </th>
</thead>

<tr bgcolor="#e5e5e5">
<td> $mean </td>
<td> $median </td>
<td> $variance </td>
<td> $temp[$#temp] </td>
<td> $temp[0] </td>
<td> $hosts{$host}{'data'}{$metric}{'start_time'} </td>
<td> $hosts{$host}{'data'}{$metric}{'stop_time'} </td>
</tr>
</table>

FIN
;

	if ($metric eq "udp_error") {
		my %data;
		my %cols;
		my $elem;
		my $lines=0;
		my $i;

		# print html table header
		print HFILE "<p></p>\n";
		print HFILE "<table border=\"0\" align=\"center\">\n";
		print HFILE "<thead>\n";
		print HFILE "<tr align=\"center\" bgcolor=\"#c0c0c0\">\n";

		foreach $elem ( @{ $fieldheaders{$metric} } ) {
       			print HFILE "<th> $elem </th>\n";
       		}
		# add start/stop time to header
#		print HFILE "<th> start </th>\n";
#		print HFILE "<th> stop </th>\n";
		print HFILE "</tr>\n";
		print HFILE "</thead>\n";
		print HFILE "<tbody>\n";

		# copy arrays
		foreach $elem ( @{ $fieldheaders{$metric} } ) {
			@{ $data{$elem} } = 
				 @{ $hosts{$host}{'data'}{$metric}{$elem} };
		}

		$lines = 0;
		while( $lines < $tablelines ) {

			if ( $lines++ % 2 == 0 ) {
		       		print HFILE "<tr bgcolor=\"#e5e5e5\">\n"
			} else {
       				print HFILE "<tr bgcolor=\"#c0c0f0\">\n"
			}

			foreach  $elem ( @{ $fieldheaders{$metric} } ) {
				$i = shift @{ $data{$elem} };
				next if ( ! defined $i );
				print HFILE "<td> $i </td>\n";

			}
			print HFILE "</tr>\n";

		}
		print HFILE "</tbody>\n";
		print HFILE "</table>\n";
	}
}


#
# prints flow html table in host html file
#
sub print_flows_table() {
	my $host = shift;
	my $flow = shift;
	my $k;
	my @row;
	my %flows;

	# flows are all the same for the moment
	# src port : dest ip . port : count

	print HFILE "<h3 align=\"left\"> $field </h3>\n";
	print HFILE "<p> Count $hosts{$host}{$flow} </p>\n";
	# process tuples keyed on dest

	foreach $k ( @{ $hosts{$host}{'data'}{$flow}{'tuple'} } ) {
		@row = split /:/, $k;
		shift @row; 	# remove leading element. kludge.
		$flows{$row[1]}{'port'} = $row[0];

		if ( ! defined $flows{$row[1]}{'count'} ) {
			$flows{$row[1]}{'count'} = 0;
		}

		$flows{$row[1]}{'count'} += $row[2];
	}

	print HFILE "<table border=\"0\" align=\"center\" >\n";
	print HFILE "<thead>\n";
	print HFILE "<tr align=\"center\" bgcolor=\"#c0c0c0\">\n";

	foreach $k ( @{ $fieldheaders{$flow} } ) {
		print HFILE "<th> $k </th>\n";
	}

	# add start/stop time to header
	print HFILE "<th> start </th>\n";
	print HFILE "<th> stop </th>\n";
	print HFILE "</tr>\n";
	print HFILE "</thead>\n";
	print HFILE "<tbody>\n";
	my $lines = 0;

#	foreach $k ( @{ $hosts{$host}{'data'}{$flow}{'tuple'} } ) {
	foreach $k ( keys %flows ) {

		# alternate row colors for readability
		if ( $line++ % 2 == 0 ) {
			print HFILE "<tr bgcolor=\"#e5e5e5\">\n";
		} else {
			print HFILE "<tr bgcolor=\"#c0c0f0\">\n"
		}

#		@row = split /:/, $k;
#		shift @row; 	# remove leading element. kludge.

#		foreach $elem (@row) {
#			print HFILE "<td> $elem </td>\n";
#		}


		print HFILE "<td> $flows{$k}{'port'} </td>\n";
		print HFILE "<td> $k </td>\n";
		print HFILE "<td> $flows{$k}{'count'} </td>\n";

		# append start/stop time
		print HFILE "<td> $hosts{$host}{'data'}{$flow}{$k}{'start_time'} </td>\n";
		print HFILE "<td> $hosts{$host}{'data'}{$flow}{$k}{'stop_time'} </td>\n";
		

		print HFILE "</tr>\n";
		$lines++;
		if ($lines > $tablelines) { last; }
	}
	print HFILE "</tbody>\n</table>\n\n";
}



#
# top html output
#
sub print_top() {

	my $date = localtime time;
	my $n = $numlines;
	my $i, $j, $k;

	open (TFILE, "> $tophtml") or die "Couldn't open $tophtml: $!\n";

	print TFILE "<html>\n";
	print TFILE "<style type=\"text/css\">\n";
	print TFILE "body, table, th, td {\n";
	print TFILE "    font-family: Arial, Helvetica, sans-serif;\n";
	print TFILE "}\n";
	print TFILE "th { text-align: center }\n";
	print TFILE "td { text-align: right; width:\"10%\" }\n";
	print TFILE "</style>\n";
	print TFILE "<body bgcolor=\"#ffffff\">\n";
	print TFILE "<head>\n";
	print TFILE "<title> top activity  </title>\n";
	print TFILE "</head>\n";
	print TFILE "<h3 align=\"left\">top catagories</h3>\n";
	print TFILE "<p align=\"left\">$date</p>\n";

	foreach $j ( @catagories ) {
		$n = numlines;
		$lines = 0;
		print TFILE "<h3 align=\"left\"> $j </h3>\n";
		print TFILE "<table border=\"0\" align=\"center\" >\n";
		print TFIEL "<thead>\n";
		print TFILE "<tr align=\"center\" bgcolor=\"#c0c0c0\"\n>";
		print TFILE "<th colspan=\"3\"></th>\n";

		#
		# define table column headings
		#
		printf TFILE "<th colspan=\" %i \">flows</th>\n", scalar (@flows);
		printf TFILE "<th colspan=\" %i \">metrics</th>\n", scalar (@metrics);
		printf TFILE "<th colspan=\" %i \">scans</th>\n", scalar(@scans);
		printf TFILE "<th colspan=\" %i \">tcpworm</th>\n", scalar(@tcpworm);

		print TFILE "</tr>\n";
		print TFILE "<tr bgcolor=\"#c0c0c0\">\n";
		print TFILE "\t<th>ip address</th>\n";
		print TFILE "\t<th>score</th>\n";
		print TFILE "\t<th>count</th>\n";
	
		foreach $i ( @fields ) { print TFILE "\t<th>$i</th>\n"; }
		print TFILE "</tr>\n";
		print TFILE "</thead>\n";
		print TFILE "<tbody>\n";
		# sort host ip addrs by count in descending order

		foreach $i (sort { $hosts{$b}{$j} <=> $hosts{$a}{$j} } keys %hosts) {
			if ( $n == 0 ) { last; };
			$lines++;

			#
			# skip hosts with a total count less than 2
			#
			if ( ! $DEBUG ) {
				next if ( $hosts{$i}{'count'} < 2 );
			}

			if ($DEBUG && $hosts{$i}{'count'} != $hosts{$i}{'sum'} ) {
				print TFILE "<tr bgcolor=\"#ff0000\">\n";
			}
			elsif ( $lines % 2 == 0 ) {
				print TFILE "<tr bgcolor=\"#e5e5e5\">\n";
			} else {
				print TFILE "<tr bgcolor=\"#c0c0f0\">\n"
			}
	
			print TFILE "\t<td>$i</td>\n";
	
			if ( $hosts{$i}{'score'} eq "" ) { 
				$hosts{$i}{'score'} = 0; 
			}

			if ( $hosts{$i}{'count'} eq "" ) { 
				$hosts{$i}{'count'} = 0; 
			}

			print TFILE "\t<td>$hosts{$i}{'score'}</td>\n";
			print TFILE "\t<td>$hosts{$i}{'count'}</td>\n";
	
			if ( $DEBUG ) { 
				print TFILE "<td>$hosts{$i}{'sum'}</td>\n";
				foreach $j ( qw/flows metrics scans/ ) {
					print TFILE "\t<td>$hosts{$i}{$j}</td>\n";
				}
			}
	
			foreach $k ( @fields ) {
				if ( $hosts{$i}{$k} eq "" ) { 
					$hosts{$i}{$k} = 0; 
				}
				print TFILE "\t<td>$hosts{$i}{$k}</td>\n";
			}
		
			print TFILE "</tr>\n";
			$n--
		}
		print TFILE "</table>\n";
	}
	print TFILE "</body>\n";
	print TFILE "</html>\n";
	close TFILE;
}



#
# html output
#
sub print_html() {

	open(HFILE, ">$htmlfile") or die "Can't open $htmlfile file: $!\n";
	my $date = localtime time;

	print HFILE <<FIN
<html>
<style type=\"text/css\">
body, table, th, td {
    font-family: Arial, Helvetica, sans-serif;
}
th { text-align: center }
td { text-align: right; width:\"10%\" }
</style>
<body bgcolor=\"#ffffff\">
<head>
<title> topipa </title>
</head>
<h3 align=\"left\">topipa</h3>
<p align=\"left\"> $start_time to $stop_time </p>
<a href="./info.topipa.html"> Technical Descripion </a>
<table border=\"0\" align=\"center\" >
<tr align=\"center\" bgcolor=\"#c0c0c0\"\n>
<th colspan=\"3\"></th>
FIN
;
	
	if ( $DEBUG ) {
		print HFILE "<th colspan=\"4\" bgcolor=\"#ffff00\">debug</th>\n";
	}
	
	#
	# define table column headings
	#
	printf HFILE "<th colspan=\" %i \">flows</th>\n", scalar (@flows);
	printf HFILE "<th colspan=\" %i \">metrics</th>\n", scalar (@metrics);
	printf HFILE "<th colspan=\" %i \">scans</th>\n", scalar (@scans);
	printf HFILE "<th colspan=\" %i \">tcpworm</th>\n", scalar (@tcpworm);

	print HFILE "</tr>\n";
	print HFILE "<tr bgcolor=\"#c0c0c0\">\n";
	print HFILE "\t<th>ip address</th>\n";
	print HFILE "\t<th>score</th>\n";
	print HFILE "\t<th>count</th>\n";
	
	if ( $DEBUG ) { 
		print HFILE "\t<th>sum</th>\n";
		print HFILE "\t<th>flows</th>\n" ;
		print HFILE "\t<th>metrics</th>\n";
		print HFILE "\t<th>scans</th>\n";
	};
	
	foreach $i ( @fields ) { print HFILE "\t<th>$i</th>\n"; }

	print HFILE "</tr>\n";
	print HFILE "<tbody>\n";

	# sort host ip addrs by count in descending order
	foreach $i (sort { $hosts{$b}{'score'} <=> $hosts{$a}{'score'} } keys %hosts) {
		my $hostweburl = $hostwebdir . '/' . $i . ".html";
		if ( $numlines == 0 ) { last; };

		#
		# skip hosts with a total count less than 2
		#
		if ( ! $DEBUG ) {
			next if ( $hosts{$i}{'count'} < 2 );
		}

		if ($DEBUG && $hosts{$i}{'count'} != $hosts{$i}{'sum'} ) {
			print HFILE "<tr bgcolor=\"#ff0000\">\n";
		}
		elsif ( $numlines % 2 == 0 ) {
			print HFILE "<tr bgcolor=\"#e5e5e5\">\n";
		} else {
			print HFILE "<tr bgcolor=\"#c0c0f0\">\n"
		}
	
		if ( $REPORT ) {
			print HFILE "\t<td><a href=\'$hostweburl\' > $i </a></td>\n";
		} else {
			print HFILE "\t<td> $i </td>\n";
		}
	
		if ( $hosts{$i}{'score'} eq "" ) { $hosts{$i}{'score'} = 0; }
		if ( $hosts{$i}{'count'} eq "" ) { $hosts{$i}{'count'} = 0; }
		print HFILE "\t<td>$hosts{$i}{'score'}</td>\n";
		print HFILE "\t<td>$hosts{$i}{'count'}</td>\n";
	
		if ( $DEBUG ) { 
			print HFILE "<td>$hosts{$i}{'sum'}</td>\n";
			foreach $j ( qw/flows metrics scans/ ) {
				print HFILE "\t<td>$hosts{$i}{$j}</td>\n";
			}
		}
	
		foreach $j ( @fields ) {
			if ( $hosts{$i}{$j} eq "" ) { $hosts{$i}{$j} = 0; }
			print HFILE "\t<td>$hosts{$i}{$j}</td>\n";
		}
	
		print HFILE "</tr>\n";
		$numlines--
	}
	print HFILE "</table>\n";
	print HFILE "<p align=\"left\">$date</p>\n";
	print HFILE "</body>\n";
	print HFILE "</html>\n";
	close(HFILE);

	
} # end sub print_html


sub printUsageAndExit()
{
	print "Usage: $0 -n num -s \"A,B,C\" -d directory -f \"mon.lite_file0 ... mon.lite_fileN\"\n";
	print "     -h            displays usage\n";
	print "     -help	  displays usage\n";
	print "     -o	          set output file (default topipa.html)\n";
	print "     -n num	  displays num lines\n";
	print "     -s A,B,C,D    weights for scoring algorithm where\n";
	print "		   A flows, B metrics, C scans, D worms\n";
	print "     -d directory  directory of mon.lite files\n";
	print "     -f file1 file2 ... fileN list of files to read\n";
	print "     -dw directory  directory of tcpworm files\n";
	print "     -fw file1 file2 ... fileN list of tcpworm files to read\n";
	print "     -report       produce HTML host output\n";
	print "     -do	          host HTML output directory\n";
	print "     -wp	          host HTML URL path\n";
	print "     -debug	  produce debugging columns\n";
	exit 1;
}



#
# processes flow data for topn_ip, topn_tcp, topn_udp, topn_icmp
#
# e.g.
# mon.lite form: topn_udp : 994 : 207.98.64.97.10->131.252.242.78.30: 4476 :  next-flow-id: byte-count: etc ...
#
# calling args:  flows_update("topn_udp", tuple args ...)
#
sub flows_update()
{
	# labels are topn_ip, topn_tcp, topn_udp or topn_icmp
	my $label = $_[0];
	my $index = $flows_index;
	my $step = $flows_step;
	my $src;
	my $dest;
	my $port;
	my $i, $j;
 
	if ( ! defined $_[$index] ) {
		return;
	} 

	# walk the flow tuples
	while ( ! ($_[$index] eq "") ) {
		# unpack flow-id
		# extract source and destination IP addresses
		($src, $dest) = split /->/, $_[$index];
		$port = $src;
		$src =~ s/\.\d+$//;			# chop port off
		$port =~ s/^\d+\.\d+\.\d+\.\d+\.//;	# save port
		$hosts{$src}{'count'}++;
		$hosts{$src}{$label}++;
		
#		flows are different. can't use subroutine.
#		timestamp_update($src, $label);

		if ($hosts{$src}{'data'}{$label}{$dest}{'start_time'} gt 
		    $timestamp ||
		    ! defined $hosts{$src}{'data'}{$label}{$dest}{'start_time'} ) {
			$hosts{$src}{'data'}{$label}{$dest}{'start_time'} = 
			    $timestamp;
		}
		if ($hosts{$src}{'data'}{$label}{$dest}{'stop_time'} lt 
		    $timestamp ||
		    ! defined $hosts{$src}{'data'}{$label}{$dest}{'stop_time'} ) {
			$hosts{$src}{'data'}{$label}{$dest}{'stop_time'} = 
			    $timestamp;
		}

		# build tuple 
		$dest = ':' . $port . ':' . $dest . ':' . $_[$index + 1]; 
		push @{ $hosts{$src}{'data'}{$label}{'tuple'} }, "$dest";

		$index = $index + $step;
	}
}



#
# scan processing routine
#
# calling args:  scans_update("ip_scan", tuple args ...)
#
sub scans_update()
{
	# labels are syn_list and worm
	# labels are ip_scan, ip_portscan, tcp_portscan or udp_portscan
	my $label = $_[0];
	my $index = $scans_index;	# first data tuple (ip addr)
	my $step = $flows_step;		# default step, totally random
	my $i, $j;

	if ( $_[$index] == 0 ) {
		return;
	} 

	# set step corresponding to label
	if ( $label eq "ip_scan" ) { $step = $step_ip_scan }
	elsif ( $label eq "tcp_scan" ) { $step = $step_tcp_scan }
	elsif ( $label eq "udp_scan" ) { $step = $step_udp_scan }
 
	while ( ! ($_[$index] eq "") ) {
		my $host = $_[$index];		# host ip address
		$hosts{$host}{'count'}++;
		$hosts{$host}{$label}++;

		# store tuple items for host processing
		$j = 1;
		foreach my $elem ( @{ $fieldheaders{$label} } ) {
			push @{ $hosts{$host}{'data'}{$label}{$elem} },
				$_[$index + $j];
			$j++;
			if (  $j >  $step ) { last; }
		}

		timestamp_update($host,$label);
		$index = $index + $step;
	} 
}




#
# metrics processing routine
#
# syn list format:
#  ip: syn: fin: rst: tcpfrom: tcpto: icmpcount: Nports: port,count,port,count,port,count
#
# calling args:  metrics_update("syn_list", tuple args ...)
#
sub metrics_update()
{
	# labels are syn_list and worm
	# labels are ip_scan, ip_portscan, tcp_portscan or udp_portscan
	my $label = $_[0];
	my $index = $metrics_index;	# first data tuple (ip addr)
	my $step = $flows_step;	# default step, totally random
	my $i, $j;

	if ( $_[$index] == 0 ) {
		return;
	} 

	# set step corresponding to label
	if ( $label eq "icmp_error" ) { $step = $step_icmp_error }
	elsif ( $label eq "udp_error" ) { $step = $step_udp_error }
	elsif ( $label eq "syn_list" ) { $step = $step_syn_list }
#	elsif ( $label eq "worm" ) { $step = $step_worm };
 
	while ( defined $_[$index] ) {
		my $host = $_[$index];		# host ip address
		$hosts{$host}{'count'}++;
		$hosts{$host}{$label}++;

		# store tuple items for host processing
		$j = 1;
		foreach my $elem ( @{ $fieldheaders{$label} } ) {
			push @{ $hosts{$host}{'data'}{$label}{$elem} },
				$_[$index + $j];
			$j++;
			if (  $j >  $step ) { last; }
		}

		timestamp_update($host,$label);
		$index = $index + $step;
	} 
}


sub timestamp_update() {

	my $host = shift;
	my $label = shift;

	if ($hosts{$host}{'data'}{$label}{'start_time'} gt $timestamp ||
	    ! defined $hosts{$host}{'data'}{$label}{'start_time'}  ){
		$hosts{$host}{'data'}{$label}{'start_time'} = 
		    $timestamp;
	}

	if ($hosts{$host}{'data'}{$label}{'stop_time'} lt $timestamp ||
    	    ! defined $hosts{$host}{'data'}{$label}{'stop'} ){
		$hosts{$host}{'data'}{$label}{'stop_time'} = 
		    $timestamp;
	}
}

# ignoring emulehit count for now
sub parseAppFlags()
{
        my $string="";
        $appflags = hex($_[0]);
        $emulehits = $_[1];

        if ($appflags & 0x0f) {
                $string = "I";          #/* irc of some sort */
        }
        if ($appflags & 0x10) {
                $string .= "B";         #/* bittorrent */
        }
        if ($appflags & 0x20) {
                $string .= "G";         #/* GNUTELLA */
        }
        if ($appflags & 0x40) {
                $string .= "K";         #/* KAZAA */
        }
        if ($appflags & 0x80 && ($emulehits >= $emulecount)) {
                $string .= "e";         #/* edonkey */
        }
#       if ($appflags & 0x100) {
#               $string .= "m";         #/* emule */
#       }
        if ($appflags & 0x200) {
                $string .= "M";         #/* Morpheus */
        }
        if ($appflags & 0x400) {
                $string .= "H";         #/* web server based on tcp src port */
        }
        if ($appflags & 0x800) {
                $string .= "E";         #/* email server based on tcp dst port */
        }
        return($string);
}
